# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Compatibility layer for API to provide both FastAPI as well as Connexion based endpoints."""

from __future__ import annotations

from airflow.providers.edge3.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    # Just re-import the types from FastAPI and Airflow Core
    from fastapi import Body, Depends, Header, HTTPException, Path, Request, status

    from airflow.api_fastapi.common.db.common import SessionDep
    from airflow.api_fastapi.common.router import AirflowRouter
    from airflow.api_fastapi.core_api.openapi.exceptions import create_openapi_http_exception_doc

    # In Airflow 3 with AIP-72 we get workload addressed by ExecuteTask
    from airflow.executors.workloads import ExecuteTask

    def parse_command(command: str) -> ExecuteTask:
        return ExecuteTask.model_validate_json(command)
else:
    # Mock the external dependencies
    from collections.abc import Callable

    from connexion import ProblemException

    class Body:  # type: ignore[no-redef]
        def __init__(self, *_, **__):
            pass

    class Depends:  # type: ignore[no-redef]
        def __init__(self, *_, **__):
            pass

    class Header:  # type: ignore[no-redef]
        def __init__(self, *_, **__):
            pass

    class Path:  # type: ignore[no-redef]
        def __init__(self, *_, **__):
            pass

    class Request:  # type: ignore[no-redef]
        pass

    class SessionDep:  # type: ignore[no-redef]
        pass

    def create_openapi_http_exception_doc(responses_status_code: list[int]) -> dict:
        return {}

    class status:  # type: ignore[no-redef]
        HTTP_204_NO_CONTENT = 204
        HTTP_400_BAD_REQUEST = 400
        HTTP_403_FORBIDDEN = 403
        HTTP_500_INTERNAL_SERVER_ERROR = 500

    class HTTPException(ProblemException):  # type: ignore[no-redef]
        """Raise when the user does not have the required permissions."""

        def __init__(
            self,
            status: int,
            detail: str,
        ) -> None:
            from airflow.utils.docs import get_docs_url

            doc_link = get_docs_url("stable-rest-api-ref.html")
            EXCEPTIONS_LINK_MAP = {
                400: f"{doc_link}#section/Errors/BadRequest",
                403: f"{doc_link}#section/Errors/PermissionDenied",
                500: f"{doc_link}#section/Errors/Unknown",
            }
            TITLE_MAP = {
                400: "BadRequest",
                403: "PermissionDenied",
                500: "InternalServerError",
            }
            super().__init__(
                status=status,
                type=EXCEPTIONS_LINK_MAP[status],
                title=TITLE_MAP[status],
                detail=detail,
            )

        def to_response(self):
            from flask import Response

            return Response(response=self.detail, status=self.status)

    class AirflowRouter:  # type: ignore[no-redef]
        def __init__(self, *_, **__):
            pass

        def get(self, *_, **__):
            def decorator(func: Callable) -> Callable:
                return func

            return decorator

        def post(self, *_, **__):
            def decorator(func: Callable) -> Callable:
                return func

            return decorator

        def patch(self, *_, **__):
            def decorator(func: Callable) -> Callable:
                return func

            return decorator

    # In Airflow 3 with AIP-72 we get workload addressed by ExecuteTask
    # But in Airflow 2.10 it is a command line array
    ExecuteTask = list[str]  # type: ignore[assignment,misc]

    def parse_command(command: str) -> ExecuteTask:
        from ast import literal_eval

        return literal_eval(command)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Compatibility layer for Connexion API to Airflow v2.10 API routes."""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Any
from uuid import uuid4

from flask import Response, request

from airflow.exceptions import AirflowException
from airflow.providers.edge3.worker_api.auth import (
    jwt_token_authorization,
    jwt_token_authorization_rpc,
)
from airflow.providers.edge3.worker_api.datamodels import (
    EdgeJobFetched,
    JsonRpcRequest,
    PushLogsBody,
    WorkerQueuesBody,
    WorkerStateBody,
)
from airflow.providers.edge3.worker_api.routes._v2_compat import HTTPException, status
from airflow.providers.edge3.worker_api.routes.jobs import fetch, state as state_api
from airflow.providers.edge3.worker_api.routes.logs import logfile_path, push_logs
from airflow.providers.edge3.worker_api.routes.worker import register, set_state
from airflow.serialization.serialized_objects import BaseSerialization
from airflow.utils.session import NEW_SESSION, create_session, provide_session

if TYPE_CHECKING:
    from airflow.api_connexion.types import APIResponse
    from airflow.utils.state import TaskInstanceState


log = logging.getLogger(__name__)


def error_response(message: str, status: int):
    """Log the error and return the response as JSON object."""
    error_id = uuid4()
    server_message = f"{message} error_id={error_id}"
    log.exception(server_message)
    client_message = f"{message} The server side traceback may be identified with error_id={error_id}"
    return HTTPException(status, client_message)


def rpcapi_v2(body: dict[str, Any]) -> APIResponse:
    """Handle Edge Worker API `/edge_worker/v1/rpcapi` endpoint for Airflow 2.10."""
    # Note: Except the method map this _was_ a 100% copy of internal API module
    #       airflow.api_internal.endpoints.rpc_api_endpoint.internal_airflow_api()
    # As of rework for FastAPI in Airflow 3.0, this is updated and to be removed in the future.
    from airflow.api_internal.endpoints.rpc_api_endpoint import (
        # Note: This is just for compatibility with Airflow 2.10, not working for Airflow 3 / main as removed
        initialize_method_map,
    )

    try:
        if request.headers.get("Content-Type", "") != "application/json":
            raise HTTPException(status.HTTP_403_FORBIDDEN, "Expected Content-Type: application/json")
        if request.headers.get("Accept", "") != "application/json":
            raise HTTPException(status.HTTP_403_FORBIDDEN, "Expected Accept: application/json")
        auth = request.headers.get("Authorization", "")
        request_obj = JsonRpcRequest(method=body["method"], jsonrpc=body["jsonrpc"], params=body["params"])
        jwt_token_authorization_rpc(request_obj, auth)
        if request_obj.jsonrpc != "2.0":
            raise error_response("Expected jsonrpc 2.0 request.", status.HTTP_400_BAD_REQUEST)

        log.debug("Got request for %s", request_obj.method)
        methods_map = initialize_method_map()
        if request_obj.method not in methods_map:
            raise error_response(f"Unrecognized method: {request_obj.method}.", status.HTTP_400_BAD_REQUEST)

        handler = methods_map[request_obj.method]
        params = {}
        try:
            if request_obj.params:
                # Note, this is Airflow 2.10 specific, as it uses Pydantic models for serialization
                params = BaseSerialization.deserialize(request_obj.params, use_pydantic_models=True)  # type: ignore[call-arg]
        except Exception:
            raise error_response("Error deserializing parameters.", status.HTTP_400_BAD_REQUEST)

        log.debug("Calling method %s\nparams: %s", request_obj.method, params)
        try:
            # Session must be created there as it may be needed by serializer for lazy-loaded fields.
            with create_session() as session:
                output = handler(**params, session=session)
                # Note, this is Airflow 2.10 specific, as it uses Pydantic models for serialization
                output_json = BaseSerialization.serialize(output, use_pydantic_models=True)  # type: ignore[call-arg]
                log.debug(
                    "Sending response: %s", json.dumps(output_json) if output_json is not None else None
                )
        # In case of AirflowException or other selective known types, transport the exception class back to caller
        except (KeyError, AttributeError, AirflowException) as e:
            # Note, this is Airflow 2.10 specific, as it uses Pydantic models for serialization
            output_json = BaseSerialization.serialize(e, use_pydantic_models=True)  # type: ignore[call-arg]
            log.debug(
                "Sending exception response: %s", json.dumps(output_json) if output_json is not None else None
            )
        except Exception:
            raise error_response(
                f"Error executing method '{request_obj.method}'.", status.HTTP_500_INTERNAL_SERVER_ERROR
            )
        response = json.dumps(output_json) if output_json is not None else None
        return Response(response=response, headers={"Content-Type": "application/json"})
    except HTTPException as e:
        return e.to_response()  # type: ignore[attr-defined]


def jwt_token_authorization_v2(method: str, authorization: str):
    """Proxy for v2 method path handling."""
    PREFIX = "/edge_worker/v1/"
    method_path = method[method.find(PREFIX) + len(PREFIX) :] if PREFIX in method else method
    jwt_token_authorization(method_path, authorization)


@provide_session
def register_v2(worker_name: str, body: dict[str, Any], session=NEW_SESSION) -> Any:
    """Handle Edge Worker API `/edge_worker/v1/worker/{worker_name}` endpoint for Airflow 2.10."""
    try:
        auth = request.headers.get("Authorization", "")
        jwt_token_authorization_v2(request.path, auth)
        request_obj = WorkerStateBody(
            state=body["state"], jobs_active=0, queues=body["queues"], sysinfo=body["sysinfo"]
        )
        return register(worker_name, request_obj, session).model_dump()
    except HTTPException as e:
        return e.to_response()  # type: ignore[attr-defined]


@provide_session
def set_state_v2(worker_name: str, body: dict[str, Any], session=NEW_SESSION) -> Any:
    """Handle Edge Worker API `/edge_worker/v1/worker/{worker_name}` endpoint for Airflow 2.10."""
    try:
        auth = request.headers.get("Authorization", "")
        jwt_token_authorization_v2(request.path, auth)
        request_obj = WorkerStateBody(
            state=body["state"],
            jobs_active=body["jobs_active"],
            queues=body["queues"],
            sysinfo=body["sysinfo"],
            maintenance_comments=body.get("maintenance_comments"),
        )
        return set_state(worker_name, request_obj, session).model_dump()
    except HTTPException as e:
        return e.to_response()  # type: ignore[attr-defined]


@provide_session
def job_fetch_v2(worker_name: str, body: dict[str, Any], session=NEW_SESSION) -> Any:
    """Handle Edge Worker API `/edge_worker/v1/jobs/fetch/{worker_name}` endpoint for Airflow 2.10."""
    from flask import request

    try:
        auth = request.headers.get("Authorization", "")
        jwt_token_authorization_v2(request.path, auth)
        queues = body.get("queues")
        free_concurrency = body.get("free_concurrency", 1)
        request_obj = WorkerQueuesBody(queues=queues, free_concurrency=free_concurrency)
        job: EdgeJobFetched | None = fetch(worker_name, request_obj, session)
        return job.model_dump() if job is not None else None
    except HTTPException as e:
        return e.to_response()  # type: ignore[attr-defined]


@provide_session
def job_state_v2(
    dag_id: str,
    task_id: str,
    run_id: str,
    try_number: int,
    map_index: str,  # Note: Connexion can not have negative numbers in path parameters, use string therefore
    state: TaskInstanceState,
    session=NEW_SESSION,
) -> Any:
    """Handle Edge Worker API `/jobs/state/{dag_id}/{task_id}/{run_id}/{try_number}/{map_index}/{state}` endpoint for Airflow 2.10."""
    from flask import request

    try:
        auth = request.headers.get("Authorization", "")
        jwt_token_authorization_v2(request.path, auth)
        state_api(dag_id, task_id, run_id, try_number, int(map_index), state, session)
    except HTTPException as e:
        return e.to_response()  # type: ignore[attr-defined]


def logfile_path_v2(
    dag_id: str,
    task_id: str,
    run_id: str,
    try_number: int,
    map_index: str,  # Note: Connexion can not have negative numbers in path parameters, use string therefore
) -> str:
    """Handle Edge Worker API `/edge_worker/v1/logs/logfile_path/{dag_id}/{task_id}/{run_id}/{try_number}/{map_index}` endpoint for Airflow 2.10."""
    try:
        auth = request.headers.get("Authorization", "")
        jwt_token_authorization_v2(request.path, auth)
        return logfile_path(dag_id, task_id, run_id, try_number, int(map_index))
    except HTTPException as e:
        return e.to_response()  # type: ignore[attr-defined]


def push_logs_v2(
    dag_id: str,
    task_id: str,
    run_id: str,
    try_number: int,
    map_index: str,  # Note: Connexion can not have negative numbers in path parameters, use string therefore
    body: dict[str, Any],
) -> None:
    """Handle Edge Worker API `/edge_worker/v1/logs/push/{dag_id}/{task_id}/{run_id}/{try_number}/{map_index}` endpoint for Airflow 2.10."""
    try:
        auth = request.headers.get("Authorization", "")
        jwt_token_authorization_v2(request.path, auth)
        request_obj = PushLogsBody(
            log_chunk_data=body["log_chunk_data"], log_chunk_time=body["log_chunk_time"]
        )
        with create_session() as session:
            push_logs(dag_id, task_id, run_id, try_number, int(map_index), request_obj, session)
    except HTTPException as e:
        return e.to_response()  # type: ignore[attr-defined]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
import os
import signal
import sys
from datetime import datetime
from http import HTTPStatus
from multiprocessing import Process
from pathlib import Path
from subprocess import Popen
from time import sleep
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from lockfile.pidlockfile import remove_existing_pidfile
from requests import HTTPError

from airflow import __version__ as airflow_version
from airflow.configuration import conf
from airflow.providers.edge3 import __version__ as edge_provider_version
from airflow.providers.edge3.cli.api_client import (
    jobs_fetch,
    jobs_set_state,
    logs_logfile_path,
    logs_push,
    worker_register,
    worker_set_state,
)
from airflow.providers.edge3.cli.dataclasses import Job, MaintenanceMarker, WorkerStatus
from airflow.providers.edge3.cli.signalling import (
    SIG_STATUS,
    maintenance_marker_file_path,
    status_file_path,
    write_pid_to_pidfile,
)
from airflow.providers.edge3.models.edge_worker import EdgeWorkerState, EdgeWorkerVersionException
from airflow.providers.edge3.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.utils import timezone
from airflow.utils.net import getfqdn
from airflow.utils.state import TaskInstanceState

if TYPE_CHECKING:
    from airflow.providers.edge3.worker_api.datamodels import EdgeJobFetched

logger = logging.getLogger(__name__)

if sys.platform == "darwin":
    setproctitle = lambda title: logger.debug("Mac OS detected, skipping setproctitle")
else:
    from setproctitle import setproctitle


def _edge_hostname() -> str:
    """Get the hostname of the edge worker that should be reported by tasks."""
    return os.environ.get("HOSTNAME", getfqdn())


class EdgeWorker:
    """Runner instance which executes the Edge Worker."""

    jobs: list[Job] = []
    """List of jobs that the worker is running currently."""
    last_hb: datetime | None = None
    """Timestamp of last heart beat sent to server."""
    drain: bool = False
    """Flag if job processing should be completed and no new jobs fetched for a graceful stop/shutdown."""
    maintenance_mode: bool = False
    """Flag if job processing should be completed and no new jobs fetched for maintenance mode. """
    maintenance_comments: str | None = None
    """Comments for maintenance mode."""
    edge_instance: EdgeWorker | None = None
    """Singleton instance of the worker."""

    def __init__(
        self,
        pid_file_path: str,
        hostname: str,
        queues: list[str] | None,
        concurrency: int,
        job_poll_interval: int,
        heartbeat_interval: int,
        daemon: bool = False,
    ):
        self.pid_file_path = pid_file_path
        self.job_poll_interval = job_poll_interval
        self.hb_interval = heartbeat_interval
        self.hostname = hostname
        self.queues = queues
        self.concurrency = concurrency
        self.free_concurrency = concurrency
        self.daemon = daemon

        EdgeWorker.edge_instance = self

    @staticmethod
    def signal_handler(sig: signal.Signals, frame):
        if sig == SIG_STATUS:
            marker_path = Path(maintenance_marker_file_path(None))
            if marker_path.exists():
                request = MaintenanceMarker.from_json(marker_path.read_text())
                logger.info("Requested to set maintenance mode to %s", request.maintenance)
                EdgeWorker.maintenance_mode = request.maintenance == "on"
                if EdgeWorker.maintenance_mode and request.comments:
                    logger.info("Comments: %s", request.comments)
                    EdgeWorker.maintenance_comments = request.comments
                marker_path.unlink()
                # send heartbeat immediately to update state
                if EdgeWorker.edge_instance:
                    EdgeWorker.edge_instance.heartbeat(EdgeWorker.maintenance_comments)
            else:
                logger.info("Request to get status of Edge Worker received.")
            status_path = Path(status_file_path(None))
            status_path.write_text(
                WorkerStatus(
                    job_count=len(EdgeWorker.jobs),
                    jobs=[job.edge_job.key for job in EdgeWorker.jobs],
                    state=EdgeWorker._get_state(),
                    maintenance=EdgeWorker.maintenance_mode,
                    maintenance_comments=EdgeWorker.maintenance_comments,
                    drain=EdgeWorker.drain,
                ).json
            )
        else:
            logger.info("Request to shut down Edge Worker received, waiting for jobs to complete.")
            EdgeWorker.drain = True

    def shutdown_handler(self, sig, frame):
        logger.info("SIGTERM received. Terminating all jobs and quit")
        for job in EdgeWorker.jobs:
            os.killpg(job.process.pid, signal.SIGTERM)
        EdgeWorker.drain = True

    def _get_sysinfo(self) -> dict:
        """Produce the sysinfo from worker to post to central site."""
        return {
            "airflow_version": airflow_version,
            "edge_provider_version": edge_provider_version,
            "concurrency": self.concurrency,
            "free_concurrency": self.free_concurrency,
        }

    @staticmethod
    def _get_state() -> EdgeWorkerState:
        """State of the Edge Worker."""
        if EdgeWorker.jobs:
            if EdgeWorker.drain:
                return EdgeWorkerState.TERMINATING
            if EdgeWorker.maintenance_mode:
                return EdgeWorkerState.MAINTENANCE_PENDING
            return EdgeWorkerState.RUNNING

        if EdgeWorker.drain:
            if EdgeWorker.maintenance_mode:
                return EdgeWorkerState.OFFLINE_MAINTENANCE
            return EdgeWorkerState.OFFLINE

        if EdgeWorker.maintenance_mode:
            return EdgeWorkerState.MAINTENANCE_MODE
        return EdgeWorkerState.IDLE

    @staticmethod
    def _run_job_via_supervisor(workload) -> int:
        from airflow.sdk.execution_time.supervisor import supervise

        # Ignore ctrl-c in this process -- we don't want to kill _this_ one. we let tasks run to completion
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        logger.info("Worker starting up pid=%d", os.getpid())
        setproctitle(f"airflow edge worker: {workload.ti.key}")

        try:
            api_url = conf.get("edge", "api_url")
            execution_api_server_url = conf.get("core", "execution_api_server_url", fallback="")
            if not execution_api_server_url:
                parsed = urlparse(api_url)
                execution_api_server_url = f"{parsed.scheme}://{parsed.netloc}/execution/"

            supervise(
                # This is the "wrong" ti type, but it duck types the same. TODO: Create a protocol for this.
                # Same like in airflow/executors/local_executor.py:_execute_work()
                ti=workload.ti,  # type: ignore[arg-type]
                dag_rel_path=workload.dag_rel_path,
                bundle_info=workload.bundle_info,
                token=workload.token,
                server=execution_api_server_url,
                log_path=workload.log_path,
            )
            return 0
        except Exception as e:
            logger.exception("Task execution failed: %s", e)
            return 1

    @staticmethod
    def _launch_job_af3(edge_job: EdgeJobFetched) -> tuple[Process, Path]:
        if TYPE_CHECKING:
            from airflow.executors.workloads import ExecuteTask

        workload: ExecuteTask = edge_job.command
        process = Process(
            target=EdgeWorker._run_job_via_supervisor,
            kwargs={"workload": workload},
        )
        process.start()
        base_log_folder = conf.get("logging", "base_log_folder", fallback="NOT AVAILABLE")
        if TYPE_CHECKING:
            assert workload.log_path  # We need to assume this is defined in here
        logfile = Path(base_log_folder, workload.log_path)
        return process, logfile

    @staticmethod
    def _launch_job_af2_10(edge_job: EdgeJobFetched) -> tuple[Popen, Path]:
        """Compatibility for Airflow 2.10 Launch."""
        env = os.environ.copy()
        env["AIRFLOW__CORE__DATABASE_ACCESS_ISOLATION"] = "True"
        env["AIRFLOW__CORE__INTERNAL_API_URL"] = conf.get("edge", "api_url")
        env["_AIRFLOW__SKIP_DATABASE_EXECUTOR_COMPATIBILITY_CHECK"] = "1"
        command: list[str] = edge_job.command  # type: ignore[assignment]
        process = Popen(command, close_fds=True, env=env, start_new_session=True)
        logfile = logs_logfile_path(edge_job.key)
        return process, logfile

    @staticmethod
    def _launch_job(edge_job: EdgeJobFetched):
        """Get the received job executed."""
        process: Popen | Process
        if AIRFLOW_V_3_0_PLUS:
            process, logfile = EdgeWorker._launch_job_af3(edge_job)
        else:
            # Airflow 2.10
            process, logfile = EdgeWorker._launch_job_af2_10(edge_job)
        EdgeWorker.jobs.append(Job(edge_job, process, logfile, 0))

    def start(self):
        """Start the execution in a loop until terminated."""
        try:
            self.last_hb = worker_register(
                self.hostname, EdgeWorkerState.STARTING, self.queues, self._get_sysinfo()
            ).last_update
        except EdgeWorkerVersionException as e:
            logger.info("Version mismatch of Edge worker and Core. Shutting down worker.")
            raise SystemExit(str(e))
        except HTTPError as e:
            if e.response.status_code == HTTPStatus.NOT_FOUND:
                raise SystemExit("Error: API endpoint is not ready, please set [edge] api_enabled=True.")
            raise SystemExit(str(e))
        if not self.daemon:
            write_pid_to_pidfile(self.pid_file_path)
        signal.signal(signal.SIGINT, EdgeWorker.signal_handler)
        signal.signal(SIG_STATUS, EdgeWorker.signal_handler)
        signal.signal(signal.SIGTERM, self.shutdown_handler)
        os.environ["HOSTNAME"] = self.hostname
        os.environ["AIRFLOW__CORE__HOSTNAME_CALLABLE"] = f"{_edge_hostname.__module__}._edge_hostname"
        try:
            self.worker_state_changed = self.heartbeat()
            self.last_hb = datetime.now()
            while not EdgeWorker.drain or EdgeWorker.jobs:
                self.loop()

            logger.info("Quitting worker, signal being offline.")
            try:
                worker_set_state(
                    self.hostname,
                    EdgeWorkerState.OFFLINE_MAINTENANCE
                    if EdgeWorker.maintenance_mode
                    else EdgeWorkerState.OFFLINE,
                    0,
                    self.queues,
                    self._get_sysinfo(),
                )
            except EdgeWorkerVersionException:
                logger.info("Version mismatch of Edge worker and Core. Quitting worker anyway.")
        finally:
            if not self.daemon:
                remove_existing_pidfile(self.pid_file_path)

    def loop(self):
        """Run a loop of scheduling and monitoring tasks."""
        new_job = False
        previous_jobs = EdgeWorker.jobs
        if not any((EdgeWorker.drain, EdgeWorker.maintenance_mode)) and self.free_concurrency > 0:
            new_job = self.fetch_job()
        self.check_running_jobs()

        if (
            EdgeWorker.drain
            or datetime.now().timestamp() - self.last_hb.timestamp() > self.hb_interval
            or self.worker_state_changed  # send heartbeat immediately if the state is different in db
            or bool(previous_jobs) != bool(EdgeWorker.jobs)  # when number of jobs changes from/to 0
        ):
            self.worker_state_changed = self.heartbeat()
            self.last_hb = datetime.now()

        if not new_job:
            self.interruptible_sleep()

    def fetch_job(self) -> bool:
        """Fetch and start a new job from central site."""
        logger.debug("Attempting to fetch a new job...")
        edge_job = jobs_fetch(self.hostname, self.queues, self.free_concurrency)
        if edge_job:
            logger.info("Received job: %s", edge_job)
            EdgeWorker._launch_job(edge_job)
            jobs_set_state(edge_job.key, TaskInstanceState.RUNNING)
            return True

        logger.info(
            "No new job to process%s",
            f", {len(EdgeWorker.jobs)} still running" if EdgeWorker.jobs else "",
        )
        return False

    def check_running_jobs(self) -> None:
        """Check which of the running tasks/jobs are completed and report back."""
        used_concurrency = 0
        for i in range(len(EdgeWorker.jobs) - 1, -1, -1):
            job = EdgeWorker.jobs[i]
            if not job.is_running:
                EdgeWorker.jobs.remove(job)
                if job.is_success:
                    logger.info("Job completed: %s", job.edge_job)
                    jobs_set_state(job.edge_job.key, TaskInstanceState.SUCCESS)
                else:
                    logger.error("Job failed: %s", job.edge_job)
                    jobs_set_state(job.edge_job.key, TaskInstanceState.FAILED)
            else:
                used_concurrency += job.edge_job.concurrency_slots

            if job.logfile.exists() and job.logfile.stat().st_size > job.logsize:
                with job.logfile.open("rb") as logfile:
                    push_log_chunk_size = conf.getint("edge", "push_log_chunk_size")
                    logfile.seek(job.logsize, os.SEEK_SET)
                    read_data = logfile.read()
                    job.logsize += len(read_data)
                    # backslashreplace to keep not decoded characters and not raising exception
                    # replace null with question mark to fix issue during DB push
                    log_data = read_data.decode(errors="backslashreplace").replace("\x00", "\ufffd")
                    while True:
                        chunk_data = log_data[:push_log_chunk_size]
                        log_data = log_data[push_log_chunk_size:]
                        if not chunk_data:
                            break

                        logs_push(
                            task=job.edge_job.key,
                            log_chunk_time=timezone.utcnow(),
                            log_chunk_data=chunk_data,
                        )

        self.free_concurrency = self.concurrency - used_concurrency

    def heartbeat(self, new_maintenance_comments: str | None = None) -> bool:
        """Report liveness state of worker to central site with stats."""
        state = EdgeWorker._get_state()
        sysinfo = self._get_sysinfo()
        worker_state_changed: bool = False
        try:
            worker_info = worker_set_state(
                self.hostname,
                state,
                len(EdgeWorker.jobs),
                self.queues,
                sysinfo,
                new_maintenance_comments,
            )
            self.queues = worker_info.queues
            if worker_info.state == EdgeWorkerState.MAINTENANCE_REQUEST:
                logger.info("Maintenance mode requested!")
                EdgeWorker.maintenance_mode = True
            elif (
                worker_info.state in [EdgeWorkerState.IDLE, EdgeWorkerState.RUNNING]
                and EdgeWorker.maintenance_mode
            ):
                logger.info("Exit Maintenance mode requested!")
                EdgeWorker.maintenance_mode = False
            if EdgeWorker.maintenance_mode:
                EdgeWorker.maintenance_comments = worker_info.maintenance_comments
            else:
                EdgeWorker.maintenance_comments = None
            if worker_info.state == EdgeWorkerState.SHUTDOWN_REQUEST:
                logger.info("Shutdown requested!")
                EdgeWorker.drain = True

            worker_state_changed = worker_info.state != state
        except EdgeWorkerVersionException:
            logger.info("Version mismatch of Edge worker and Core. Shutting down worker.")
            EdgeWorker.drain = True
        return worker_state_changed

    def interruptible_sleep(self):
        """Sleeps but stops sleeping if drain is made."""
        drain_before_sleep = EdgeWorker.drain
        for _ in range(0, self.job_poll_interval * 10):
            sleep(0.1)
            if drain_before_sleep != EdgeWorker.drain:
                return



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException
from airflow.plugins_manager import AirflowPlugin
from airflow.providers.edge3.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_1_PLUS
from airflow.utils.session import NEW_SESSION, provide_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

if AIRFLOW_V_3_0_PLUS:
    from airflow.utils.db import DBLocks, create_global_lock

    @provide_session
    def _get_api_endpoint(session: Session = NEW_SESSION) -> dict[str, Any]:
        # Ensure all required DB modeals are created before starting the API
        with create_global_lock(session=session, lock=DBLocks.MIGRATIONS):
            engine = session.get_bind().engine
            from airflow.providers.edge3.models.edge_job import EdgeJobModel
            from airflow.providers.edge3.models.edge_logs import EdgeLogsModel
            from airflow.providers.edge3.models.edge_worker import EdgeWorkerModel

            EdgeJobModel.metadata.create_all(engine)
            EdgeLogsModel.metadata.create_all(engine)
            EdgeWorkerModel.metadata.create_all(engine)

        from airflow.providers.edge3.worker_api.app import create_edge_worker_api_app

        return {
            "app": create_edge_worker_api_app(),
            "url_prefix": "/edge_worker",
            "name": "Airflow Edge Worker",
        }

else:
    # This is for back-compatibility with Airflow 2.x and we only make this
    # to prevents dependencies and breaking imports in Airflow 3.x
    import re
    from datetime import datetime, timedelta
    from pathlib import Path

    from flask import Blueprint, redirect, request, url_for
    from flask_appbuilder import BaseView, expose
    from markupsafe import Markup
    from sqlalchemy import select

    from airflow.auth.managers.models.resource_details import AccessView
    from airflow.models.taskinstance import TaskInstanceState
    from airflow.utils.state import State
    from airflow.utils.yaml import safe_load
    from airflow.www.auth import has_access_view

    def _get_airflow_2_api_endpoint() -> Blueprint:
        from airflow.www.app import csrf
        from airflow.www.constants import SWAGGER_BUNDLE, SWAGGER_ENABLED
        from airflow.www.extensions.init_views import _CustomErrorRequestBodyValidator, _LazyResolver

        folder = Path(__file__).parents[1].resolve()  # this is airflow/providers/edge3/
        with folder.joinpath("openapi", "edge_worker_api_v1.yaml").open() as f:
            specification = safe_load(f)
        from connexion import FlaskApi

        bp = FlaskApi(
            specification=specification,
            resolver=_LazyResolver(),
            base_path="/edge_worker/v1",
            strict_validation=True,
            options={"swagger_ui": SWAGGER_ENABLED, "swagger_path": SWAGGER_BUNDLE.__fspath__()},
            validate_responses=True,
            validator_map={"body": _CustomErrorRequestBodyValidator},
        ).blueprint
        # Need to exempt CSRF to make API usable
        csrf.exempt(bp)
        return bp

    def _state_token(state):
        """Return a formatted string with HTML for a given State."""
        color = State.color(state)
        fg_color = State.color_fg(state)
        return Markup(
            """
            <span class="label" style="color:{fg_color}; background-color:{color};"
                title="Current State: {state}">{state}</span>
            """
        ).format(color=color, state=state, fg_color=fg_color)

    def modify_maintenance_comment_on_update(maintenance_comment: str | None, username: str) -> str:
        if maintenance_comment:
            if re.search(
                r"^\[[-\d:\s]+\] - .+ put node into maintenance mode\r?\nComment:.*", maintenance_comment
            ):
                return re.sub(
                    r"^\[[-\d:\s]+\] - .+ put node into maintenance mode\r?\nComment:",
                    f"[{datetime.now().strftime('%Y-%m-%d %H:%M')}] - {username} updated maintenance mode\nComment:",
                    maintenance_comment,
                )
            if re.search(r"^\[[-\d:\s]+\] - .+ updated maintenance mode\r?\nComment:.*", maintenance_comment):
                return re.sub(
                    r"^\[[-\d:\s]+\] - .+ updated maintenance mode\r?\nComment:",
                    f"[{datetime.now().strftime('%Y-%m-%d %H:%M')}] - {username} updated maintenance mode\nComment:",
                    maintenance_comment,
                )
            return f"[{datetime.now().strftime('%Y-%m-%d %H:%M')}] - {username} updated maintenance mode\nComment: {maintenance_comment}"
        return (
            f"[{datetime.now().strftime('%Y-%m-%d %H:%M')}] - {username} updated maintenance mode\nComment:"
        )

    # registers airflow/providers/edge3/plugins/templates as a Jinja template folder
    template_bp = Blueprint(
        "template_blueprint",
        __name__,
        template_folder="templates",
    )

    class EdgeWorkerJobs(BaseView):
        """Simple view to show Edge Worker jobs."""

        default_view = "jobs"

        @expose("/jobs")
        @has_access_view(AccessView.JOBS)
        @provide_session
        def jobs(self, session: Session = NEW_SESSION):
            from airflow.providers.edge3.models.edge_job import EdgeJobModel

            jobs = session.scalars(select(EdgeJobModel).order_by(EdgeJobModel.queued_dttm)).all()
            html_states = {
                str(state): _state_token(str(state)) for state in TaskInstanceState.__members__.values()
            }
            return self.render_template("edge_worker_jobs.html", jobs=jobs, html_states=html_states)

    class EdgeWorkerHosts(BaseView):
        """Simple view to show Edge Worker status."""

        default_view = "status"

        @expose("/status")
        @has_access_view(AccessView.JOBS)
        @provide_session
        def status(self, session: Session = NEW_SESSION):
            from airflow.providers.edge3.models.edge_worker import EdgeWorkerModel

            hosts = session.scalars(select(EdgeWorkerModel).order_by(EdgeWorkerModel.worker_name)).all()
            five_min_ago = datetime.now() - timedelta(minutes=5)
            return self.render_template("edge_worker_hosts.html", hosts=hosts, five_min_ago=five_min_ago)

        @expose("/status/maintenance/<string:worker_name>/on", methods=["POST"])
        @has_access_view(AccessView.JOBS)
        def worker_to_maintenance(self, worker_name: str):
            from flask_login import current_user

            from airflow.providers.edge3.models.edge_worker import request_maintenance

            maintenance_comment = request.form.get("maintenance_comment")
            maintenance_comment = f"[{datetime.now().strftime('%Y-%m-%d %H:%M')}] - {current_user.username} put node into maintenance mode\nComment: {maintenance_comment}"
            request_maintenance(worker_name, maintenance_comment)
            return redirect(url_for("EdgeWorkerHosts.status"))

        @expose("/status/maintenance/<string:worker_name>/off", methods=["POST"])
        @has_access_view(AccessView.JOBS)
        def remove_worker_from_maintenance(self, worker_name: str):
            from airflow.providers.edge3.models.edge_worker import exit_maintenance

            exit_maintenance(worker_name)
            return redirect(url_for("EdgeWorkerHosts.status"))

        @expose("/status/maintenance/<string:worker_name>/remove", methods=["POST"])
        @has_access_view(AccessView.JOBS)
        def remove_worker(self, worker_name: str):
            from airflow.providers.edge3.models.edge_worker import remove_worker

            remove_worker(worker_name)
            return redirect(url_for("EdgeWorkerHosts.status"))

        @expose("/status/maintenance/<string:worker_name>/change_comment", methods=["POST"])
        @has_access_view(AccessView.JOBS)
        def change_maintenance_comment(self, worker_name: str):
            from flask_login import current_user

            from airflow.providers.edge3.models.edge_worker import change_maintenance_comment

            maintenance_comment = request.form.get("maintenance_comment")
            maintenance_comment = modify_maintenance_comment_on_update(
                maintenance_comment, current_user.username
            )
            change_maintenance_comment(worker_name, maintenance_comment)
            return redirect(url_for("EdgeWorkerHosts.status"))


# Check if EdgeExecutor is actually loaded
try:
    EDGE_EXECUTOR_ACTIVE = conf.getboolean("edge", "api_enabled", fallback="False")
except AirflowConfigException:
    EDGE_EXECUTOR_ACTIVE = False

# Load the API endpoint only on api-server (Airflow 3.x) or webserver (Airflow 2.x)
# todo(jscheffl): Remove this check when the discussion in
#                 https://lists.apache.org/thread/w170czq6r7bslkqp1tk6bjjjo0789wgl
#                 resulted in a proper API to selective initialize. Maybe backcompat-shim
#                 is also needed to support Airflow-versions prior the rework.
if AIRFLOW_V_3_0_PLUS:
    RUNNING_ON_APISERVER = (len(sys.argv) > 1 and sys.argv[1] in ["api-server"]) or (
        len(sys.argv) > 2 and sys.argv[2] == "airflow-core/src/airflow/api_fastapi/main.py"
    )
else:
    RUNNING_ON_APISERVER = "gunicorn" in sys.argv[0] and "airflow-webserver" in sys.argv


class EdgeExecutorPlugin(AirflowPlugin):
    """EdgeExecutor Plugin - provides API endpoints for Edge Workers in Webserver."""

    name = "edge_executor"
    if EDGE_EXECUTOR_ACTIVE and RUNNING_ON_APISERVER:
        if AIRFLOW_V_3_1_PLUS:
            fastapi_apps = [_get_api_endpoint()]
            react_apps = [
                {
                    "name": "Edge Worker",
                    "bundle_url": "/edge_worker/static/main.umd.cjs",
                    "destination": "nav",
                    "url_route": "edge_worker",
                    "category": "admin",
                    "icon": "/edge_worker/res/cloud-computer.svg",
                    "icon_dark_mode": "/edge_worker/res/cloud-computer-dark.svg",
                },
                {
                    "name": "Edge Worker Jobs",
                    "bundle_url": "/edge_worker/static/main.umd.cjs",
                    "url_route": "edge_jobs",
                    "category": "admin",
                    "icon": "/edge_worker/res/cloud-computer.svg",
                    "icon_dark_mode": "/edge_worker/res/cloud-computer-dark.svg",
                },
            ]
            external_views = [
                {
                    "name": "Edge Worker API docs",
                    "href": "/edge_worker/docs",
                    "destination": "nav",
                    "category": "docs",
                    "icon": "/edge_worker/res/cloud-computer.svg",
                    "icon_dark_mode": "/edge_worker/res/cloud-computer-dark.svg",
                    "url_route": "edge_worker_api_docs",
                }
            ]
        if AIRFLOW_V_3_0_PLUS:
            # Airflow 3.0 does not know about react_apps, so we only provide the API endpoint
            fastapi_apps = [_get_api_endpoint()]
        else:
            appbuilder_menu_items = [
                {
                    "name": "Edge Worker API docs",
                    "href": "/edge_worker/v1/ui",
                    "category": "Docs",
                }
            ]
            appbuilder_views = [
                {
                    "name": "Edge Worker Jobs",
                    "category": "Admin",
                    "view": EdgeWorkerJobs(),
                },
                {
                    "name": "Edge Worker Hosts",
                    "category": "Admin",
                    "view": EdgeWorkerHosts(),
                },
            ]
            flask_blueprints = [_get_airflow_2_api_endpoint(), template_bp]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from copy import deepcopy
from datetime import datetime, timedelta
from unittest.mock import MagicMock, patch

import pytest
import time_machine

from airflow.configuration import conf
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.providers.edge3.executors.edge_executor import EdgeExecutor
from airflow.providers.edge3.models.edge_job import EdgeJobModel
from airflow.providers.edge3.models.edge_worker import EdgeWorkerModel, EdgeWorkerState
from airflow.utils import timezone
from airflow.utils.session import create_session
from airflow.utils.state import TaskInstanceState

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

pytestmark = pytest.mark.db_test


class TestEdgeExecutor:
    @pytest.fixture(autouse=True)
    def setup_test_cases(self):
        with create_session() as session:
            session.query(EdgeJobModel).delete()

    def get_test_executor(self, pool_slots=1):
        key = TaskInstanceKey(
            dag_id="test_dag", run_id="test_run", task_id="test_task", map_index=-1, try_number=1
        )
        ti = MagicMock()
        ti.pool_slots = pool_slots
        ti.dag_run.dag_id = key.dag_id
        ti.dag_run.run_id = key.run_id
        ti.dag_run.start_date = datetime(2021, 1, 1)
        executor = EdgeExecutor()
        executor.queued_tasks = {key: [None, None, None, ti]}

        return (executor, key)

    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="_process_tasks is not used in Airflow 3.0+")
    def test__process_tasks_bad_command(self):
        executor, key = self.get_test_executor()
        task_tuple = (key, ["hello", "world"], None, None)
        with pytest.raises(ValueError):
            executor._process_tasks([task_tuple])

    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="_process_tasks is not used in Airflow 3.0+")
    @pytest.mark.parametrize(
        "pool_slots, expected_concurrency",
        [
            pytest.param(1, 1, id="default_pool_size"),
            pytest.param(5, 5, id="increased_pool_size"),
        ],
    )
    def test__process_tasks_ok_command(self, pool_slots, expected_concurrency):
        executor, key = self.get_test_executor(pool_slots=pool_slots)
        task_tuple = (key, ["airflow", "tasks", "run", "hello", "world"], None, None)
        executor._process_tasks([task_tuple])

        with create_session() as session:
            jobs: list[EdgeJobModel] = session.query(EdgeJobModel).all()
        assert len(jobs) == 1
        assert jobs[0].dag_id == "test_dag"
        assert jobs[0].run_id == "test_run"
        assert jobs[0].task_id == "test_task"
        assert jobs[0].concurrency_slots == expected_concurrency

    @patch("airflow.stats.Stats.incr")
    def test_sync_orphaned_tasks(self, mock_stats_incr):
        executor = EdgeExecutor()

        delta_to_purge = timedelta(minutes=conf.getint("edge", "job_fail_purge") + 1)
        if AIRFLOW_V_3_0_PLUS:
            delta_to_orphaned_config_name = "task_instance_heartbeat_timeout"
        else:
            delta_to_orphaned_config_name = "scheduler_zombie_task_threshold"

        delta_to_orphaned = timedelta(seconds=conf.getint("scheduler", delta_to_orphaned_config_name) + 1)

        with create_session() as session:
            for task_id, state, last_update in [
                (
                    "started_running_orphaned",
                    TaskInstanceState.RUNNING,
                    timezone.utcnow() - delta_to_orphaned,
                ),
                ("started_removed", TaskInstanceState.REMOVED, timezone.utcnow() - delta_to_purge),
            ]:
                session.add(
                    EdgeJobModel(
                        dag_id="test_dag",
                        task_id=task_id,
                        run_id="test_run",
                        map_index=-1,
                        try_number=1,
                        state=state,
                        queue="default",
                        command="mock",
                        concurrency_slots=1,
                        last_update=last_update,
                    )
                )
                session.commit()

        executor.sync()

        mock_stats_incr.assert_called_with(
            "edge_worker.ti.finish",
            tags={
                "dag_id": "test_dag",
                "queue": "default",
                "state": "failed",
                "task_id": "started_running_orphaned",
            },
        )
        mock_stats_incr.call_count == 2

        with create_session() as session:
            jobs = session.query(EdgeJobModel).all()
            assert len(jobs) == 1
            assert jobs[0].task_id == "started_running_orphaned"
            assert jobs[0].state == TaskInstanceState.REMOVED

    @patch("airflow.providers.edge3.executors.edge_executor.EdgeExecutor.running_state")
    @patch("airflow.providers.edge3.executors.edge_executor.EdgeExecutor.success")
    @patch("airflow.providers.edge3.executors.edge_executor.EdgeExecutor.fail")
    def test_sync(self, mock_fail, mock_success, mock_running_state):
        executor = EdgeExecutor()

        def remove_from_running(key: TaskInstanceKey):
            executor.running.remove(key)

        mock_success.side_effect = remove_from_running
        mock_fail.side_effect = remove_from_running

        delta_to_purge = timedelta(minutes=conf.getint("edge", "job_fail_purge") + 1)

        # Prepare some data
        with create_session() as session:
            for task_id, state, last_update in [
                ("started_running", TaskInstanceState.RUNNING, timezone.utcnow()),
                ("started_success", TaskInstanceState.SUCCESS, timezone.utcnow() - delta_to_purge),
                ("started_failed", TaskInstanceState.FAILED, timezone.utcnow() - delta_to_purge),
            ]:
                session.add(
                    EdgeJobModel(
                        dag_id="test_dag",
                        task_id=task_id,
                        run_id="test_run",
                        map_index=-1,
                        try_number=1,
                        state=state,
                        queue="default",
                        concurrency_slots=1,
                        command="mock",
                        last_update=last_update,
                    )
                )
                key = TaskInstanceKey(
                    dag_id="test_dag", run_id="test_run", task_id=task_id, map_index=-1, try_number=1
                )
                executor.running.add(key)
                session.commit()
        assert len(executor.running) == 3

        executor.sync()

        with create_session() as session:
            jobs = session.query(EdgeJobModel).all()
            assert len(session.query(EdgeJobModel).all()) == 1
            assert jobs[0].task_id == "started_running"
            assert jobs[0].state == TaskInstanceState.RUNNING

        assert len(executor.running) == 1
        mock_running_state.assert_called_once()
        mock_success.assert_called_once()
        mock_fail.assert_called_once()

        # Any test another round with one new run
        mock_running_state.reset_mock()
        mock_success.reset_mock()
        mock_fail.reset_mock()

        with create_session() as session:
            task_id = "started_running2"
            state = TaskInstanceState.RUNNING
            session.add(
                EdgeJobModel(
                    dag_id="test_dag",
                    task_id=task_id,
                    run_id="test_run",
                    map_index=-1,
                    try_number=1,
                    state=state,
                    concurrency_slots=1,
                    queue="default",
                    command="mock",
                    last_update=timezone.utcnow(),
                )
            )
            key = TaskInstanceKey(
                dag_id="test_dag", run_id="test_run", task_id=task_id, map_index=-1, try_number=1
            )
            executor.running.add(key)
            session.commit()
        assert len(executor.running) == 2

        executor.sync()

        assert len(executor.running) == 2
        mock_running_state.assert_called_once()  # because we reported already first run, new run calling
        mock_success.assert_not_called()  # because we reported already, not running anymore
        mock_fail.assert_not_called()  # because we reported already, not running anymore
        mock_running_state.reset_mock()

        executor.sync()

        assert len(executor.running) == 2
        # Now none is called as we called before already
        mock_running_state.assert_not_called()
        mock_success.assert_not_called()
        mock_fail.assert_not_called()

    def test_sync_active_worker(self):
        executor = EdgeExecutor()

        # Prepare some data
        with create_session() as session:
            # Clear existing workers to avoid unique constraint violation
            session.query(EdgeWorkerModel).delete()
            session.commit()

            # Add workers with different states
            for worker_name, state, last_heartbeat in [
                (
                    "inactive_timed_out_worker",
                    EdgeWorkerState.IDLE,
                    datetime(2023, 1, 1, 0, 59, 0, tzinfo=timezone.utc),
                ),
                ("active_worker", EdgeWorkerState.IDLE, datetime(2023, 1, 1, 0, 59, 10, tzinfo=timezone.utc)),
                (
                    "offline_worker",
                    EdgeWorkerState.OFFLINE,
                    datetime(2023, 1, 1, 0, 59, 10, tzinfo=timezone.utc),
                ),
                (
                    "offline_maintenance_worker",
                    EdgeWorkerState.OFFLINE_MAINTENANCE,
                    datetime(2023, 1, 1, 0, 59, 10, tzinfo=timezone.utc),
                ),
            ]:
                session.add(
                    EdgeWorkerModel(
                        worker_name=worker_name,
                        state=state,
                        last_update=last_heartbeat,
                        queues="",
                        first_online=timezone.utcnow(),
                    )
                )
                session.commit()

        with time_machine.travel(datetime(2023, 1, 1, 1, 0, 0, tzinfo=timezone.utc), tick=False):
            with conf_vars({("edge", "heartbeat_interval"): "10"}):
                executor.sync()

        with create_session() as session:
            for worker in session.query(EdgeWorkerModel).all():
                print(worker.worker_name)
                if "maintenance_" in worker.worker_name:
                    EdgeWorkerState.OFFLINE_MAINTENANCE
                elif "offline_" in worker.worker_name:
                    assert worker.state == EdgeWorkerState.OFFLINE
                elif "inactive_" in worker.worker_name:
                    assert worker.state == EdgeWorkerState.UNKNOWN
                else:
                    assert worker.state == EdgeWorkerState.IDLE

    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="API only available in Airflow <3.0")
    def test_execute_async(self):
        executor, key = self.get_test_executor()

        # Need to apply "trick" which is used to pass pool_slots
        executor.edge_queued_tasks = deepcopy(executor.queued_tasks)

        executor.execute_async(key=key, command=["airflow", "tasks", "run", "hello", "world"])

        with create_session() as session:
            jobs = session.query(EdgeJobModel).all()
            assert len(jobs) == 1

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="API only available in Airflow 3.0+")
    def test_queue_workload(self):
        from airflow.executors.workloads import ExecuteTask, TaskInstance

        executor = self.get_test_executor()[0]

        with pytest.raises(TypeError):
            # Does not like the Airflow 2.10 type of workload
            executor.queue_workload(command=["airflow", "tasks", "run", "hello", "world"])

        workload = ExecuteTask(
            token="mock",
            ti=TaskInstance(
                id="4d828a62-a417-4936-a7a6-2b3fabacecab",
                task_id="mock",
                dag_id="mock",
                run_id="mock",
                try_number=1,
                pool_slots=1,
                queue="default",
                priority_weight=1,
                start_date=timezone.utcnow(),
                dag_version_id="4d828a62-a417-4936-a7a6-2b3fabacecab",
            ),
            dag_rel_path="mock.py",
            log_path="mock.log",
            bundle_info={"name": "n/a", "version": "no matter"},
        )
        executor.queue_workload(workload=workload)

        with create_session() as session:
            jobs = session.query(EdgeJobModel).all()
            assert len(jobs) == 1

    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="API only available in Airflow <3.0")
    def test_execute_async_updates_existing_job(self):
        executor, key = self.get_test_executor()

        # First insert a job with the same key
        with create_session() as session:
            session.add(
                EdgeJobModel(
                    dag_id=key.dag_id,
                    run_id=key.run_id,
                    task_id=key.task_id,
                    map_index=key.map_index,
                    try_number=key.try_number,
                    state=TaskInstanceState.SCHEDULED,
                    queue="default",
                    concurrency_slots=1,
                    command="old-command",
                    last_update=timezone.utcnow(),
                )
            )
            session.commit()

        # Trigger execute_async which should update the existing job
        executor.edge_queued_tasks = deepcopy(executor.queued_tasks)
        executor.execute_async(key=key, command=["airflow", "tasks", "run", "new", "command"])

        with create_session() as session:
            jobs = session.query(EdgeJobModel).all()
            assert len(jobs) == 1
            job = jobs[0]
            assert job.state == TaskInstanceState.QUEUED
            assert job.command != "old-command"
            assert "new" in job.command

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="API only available in Airflow 3.0+")
    def test_queue_workload_updates_existing_job(self):
        from uuid import uuid4

        from airflow.executors.workloads import ExecuteTask, TaskInstance

        executor = self.get_test_executor()[0]

        key = TaskInstanceKey(dag_id="mock", run_id="mock", task_id="mock", map_index=-1, try_number=1)

        # Insert an existing job
        with create_session() as session:
            session.add(
                EdgeJobModel(
                    dag_id=key.dag_id,
                    task_id=key.task_id,
                    run_id=key.run_id,
                    map_index=key.map_index,
                    try_number=key.try_number,
                    state=TaskInstanceState.SCHEDULED,
                    queue="default",
                    command="old-command",
                    concurrency_slots=1,
                    last_update=timezone.utcnow(),
                )
            )
            session.commit()

        # Queue a workload with same key
        workload = ExecuteTask(
            token="mock",
            ti=TaskInstance(
                id=uuid4(),
                task_id=key.task_id,
                dag_id=key.dag_id,
                run_id=key.run_id,
                try_number=key.try_number,
                map_index=key.map_index,
                pool_slots=1,
                queue="updated-queue",
                priority_weight=1,
                start_date=timezone.utcnow(),
                dag_version_id=uuid4(),
            ),
            dag_rel_path="mock.py",
            log_path="mock.log",
            bundle_info={"name": "n/a", "version": "no matter"},
        )

        executor.queue_workload(workload=workload)

        with create_session() as session:
            jobs = session.query(EdgeJobModel).all()
            assert len(jobs) == 1
            job = jobs[0]
            assert job.queue == "updated-queue"
            assert job.command != "old-command"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import argparse
import contextlib
import importlib
import json
from datetime import datetime
from io import StringIO
from pathlib import Path
from subprocess import Popen
from unittest.mock import MagicMock, call, patch

import pytest
import time_machine
from requests import HTTPError, Response

from airflow.cli import cli_parser
from airflow.executors import executor_loader
from airflow.providers.edge3.cli import edge_command
from airflow.providers.edge3.cli.dataclasses import Job
from airflow.providers.edge3.cli.worker import EdgeWorker
from airflow.providers.edge3.models.edge_worker import (
    EdgeWorkerModel,
    EdgeWorkerState,
    EdgeWorkerVersionException,
)
from airflow.providers.edge3.worker_api.datamodels import (
    EdgeJobFetched,
    WorkerRegistrationReturn,
    WorkerSetStateReturn,
)
from airflow.utils import timezone
from airflow.utils.state import TaskInstanceState

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

pytest.importorskip("pydantic", minversion="2.0.0")


MOCK_COMMAND = (
    {
        "token": "mock",
        "ti": {
            "id": "4d828a62-a417-4936-a7a6-2b3fabacecab",
            "task_id": "mock",
            "dag_id": "mock",
            "run_id": "mock",
            "try_number": 1,
            "dag_version_id": "01234567-89ab-cdef-0123-456789abcdef",
            "pool_slots": 1,
            "queue": "default",
            "priority_weight": 1,
            "start_date": "2023-01-01T00:00:00+00:00",
            "map_index": -1,
        },
        "dag_rel_path": "mock.py",
        "log_path": "mock.log",
        "bundle_info": {"name": "hello", "version": "abc"},
    }
    if AIRFLOW_V_3_0_PLUS
    else ["test", "command"]  # Airflow 2.10
)


class _MockPopen(Popen):
    def __init__(self, returncode=None):
        self.generated_returncode = None

    def poll(self):
        pass

    @property
    def returncode(self):
        return self.generated_returncode


class TestEdgeWorker:
    parser: argparse.ArgumentParser

    @classmethod
    def setup_class(cls):
        with conf_vars(
            {("core", "executor"): "airflow.providers.edge3.executors.edge_executor.EdgeExecutor"}
        ):
            importlib.reload(executor_loader)
            importlib.reload(cli_parser)
            cls.parser = cli_parser.get_parser()

    @pytest.fixture
    def mock_joblist(self, tmp_path: Path) -> list[Job]:
        logfile = tmp_path / "file.log"
        logfile.touch()

        return [
            Job(
                edge_job=EdgeJobFetched(
                    dag_id="test",
                    task_id="test1",
                    run_id="test",
                    map_index=-1,
                    try_number=1,
                    concurrency_slots=1,
                    command=MOCK_COMMAND,  # type: ignore[arg-type]
                ),
                process=_MockPopen(),
                logfile=logfile,
                logsize=0,
            ),
        ]

    @pytest.fixture
    def worker_with_job(self, tmp_path: Path, mock_joblist: list[Job]) -> EdgeWorker:
        test_worker = EdgeWorker(str(tmp_path / "mock.pid"), "mock", None, 8, 5, 5)
        EdgeWorker.jobs = mock_joblist
        return test_worker

    @pytest.fixture
    def mock_edgeworker(self) -> EdgeWorkerModel:
        test_edgeworker = EdgeWorkerModel(
            worker_name="test_edge_worker",
            state="idle",
            queues=["default"],
        )
        return test_edgeworker

    @patch("airflow.providers.edge3.cli.worker.Process")
    @patch("airflow.providers.edge3.cli.worker.logs_logfile_path")
    @patch("airflow.providers.edge3.cli.worker.Popen")
    def test_launch_job(self, mock_popen, mock_logfile_path, mock_process, worker_with_job: EdgeWorker):
        mock_popen.side_effect = [MagicMock()]
        mock_process_instance = MagicMock()
        mock_process.side_effect = [mock_process_instance]

        edge_job = EdgeWorker.jobs.pop().edge_job
        with conf_vars({("edge", "api_url"): "https://invalid-api-test-endpoint"}):
            worker_with_job._launch_job(edge_job)

        if AIRFLOW_V_3_0_PLUS:
            assert mock_process.call_count == 1
            assert mock_process_instance.start.call_count == 1
        else:
            assert mock_popen.call_count == 1
            assert mock_logfile_path.call_count == 1

        assert len(EdgeWorker.jobs) == 1
        assert EdgeWorker.jobs[0].edge_job == edge_job

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3+")
    @pytest.mark.parametrize(
        "configs, expected_url",
        [
            (
                {("edge", "api_url"): "https://api-endpoint"},
                "https://api-endpoint/execution/",
            ),
            (
                {("edge", "api_url"): "https://api:1234/endpoint"},
                "https://api:1234/execution/",
            ),
            (
                {
                    ("edge", "api_url"): "https://api-endpoint",
                    ("core", "execution_api_server_url"): "https://other-endpoint",
                },
                "https://other-endpoint",
            ),
        ],
    )
    @patch("airflow.sdk.execution_time.supervisor.supervise")
    @patch("airflow.providers.edge3.cli.worker.Process")
    @patch("airflow.providers.edge3.cli.worker.Popen")
    def test_use_execution_api_server_url(
        self,
        mock_popen,
        mock_process,
        mock_supervise,
        configs,
        expected_url,
        worker_with_job: EdgeWorker,
    ):
        mock_popen.side_effect = [MagicMock()]
        mock_process_instance = MagicMock()
        mock_process.side_effect = [mock_process_instance]

        edge_job = EdgeWorker.jobs.pop().edge_job
        with conf_vars(configs):
            worker_with_job._launch_job(edge_job)

            mock_process_callback = mock_process.call_args.kwargs["target"]
            mock_process_callback(workload=MagicMock())

            assert mock_supervise.call_args.kwargs["server"] == expected_url

    @pytest.mark.parametrize(
        "reserve_result, fetch_result, expected_calls",
        [
            pytest.param(None, False, (0, 0), id="no_job"),
            pytest.param(
                EdgeJobFetched(
                    dag_id="test",
                    task_id="test",
                    run_id="test",
                    map_index=-1,
                    try_number=1,
                    concurrency_slots=1,
                    command=MOCK_COMMAND,  # type: ignore[arg-type]
                ),
                True,
                (1, 1),
                id="new_job",
            ),
        ],
    )
    @patch("airflow.providers.edge3.cli.worker.jobs_fetch")
    @patch("airflow.providers.edge3.cli.worker.logs_logfile_path")
    @patch("airflow.providers.edge3.cli.worker.jobs_set_state")
    @patch("subprocess.Popen")
    def test_fetch_job(
        self,
        mock_popen,
        mock_set_state,
        mock_logfile_path,
        mock_reserve_task,
        reserve_result,
        fetch_result,
        expected_calls,
        worker_with_job: EdgeWorker,
    ):
        logfile_path_call_count, set_state_call_count = expected_calls
        mock_reserve_task.side_effect = [reserve_result]
        mock_popen.side_effect = ["mock"]
        with conf_vars({("edge", "api_url"): "https://invalid-api-test-endpoint"}):
            got_job = worker_with_job.fetch_job()
        mock_reserve_task.assert_called_once()
        assert got_job == fetch_result
        if AIRFLOW_V_3_0_PLUS:
            # this is only called on Airflow 2.10, AIP-72 includes it
            assert mock_logfile_path.call_count == 0
        else:
            assert mock_logfile_path.call_count == logfile_path_call_count
        assert mock_set_state.call_count == set_state_call_count

    def test_check_running_jobs_running(self, worker_with_job: EdgeWorker):
        assert worker_with_job.free_concurrency == worker_with_job.concurrency
        with conf_vars({("edge", "api_url"): "https://invalid-api-test-endpoint"}):
            worker_with_job.check_running_jobs()
        assert len(EdgeWorker.jobs) == 1
        assert (
            worker_with_job.free_concurrency
            == worker_with_job.concurrency - EdgeWorker.jobs[0].edge_job.concurrency_slots
        )

    @patch("airflow.providers.edge3.cli.worker.jobs_set_state")
    def test_check_running_jobs_success(self, mock_set_state, worker_with_job: EdgeWorker):
        job = EdgeWorker.jobs[0]
        job.process.generated_returncode = 0  # type: ignore[union-attr]
        with conf_vars({("edge", "api_url"): "https://invalid-api-test-endpoint"}):
            worker_with_job.check_running_jobs()
        assert len(EdgeWorker.jobs) == 0
        mock_set_state.assert_called_once_with(job.edge_job.key, TaskInstanceState.SUCCESS)
        assert worker_with_job.free_concurrency == worker_with_job.concurrency

    @patch("airflow.providers.edge3.cli.worker.jobs_set_state")
    def test_check_running_jobs_failed(self, mock_set_state, worker_with_job: EdgeWorker):
        job = EdgeWorker.jobs[0]
        job.process.generated_returncode = 42  # type: ignore[union-attr]
        with conf_vars({("edge", "api_url"): "https://invalid-api-test-endpoint"}):
            worker_with_job.check_running_jobs()
        assert len(EdgeWorker.jobs) == 0
        mock_set_state.assert_called_once_with(job.edge_job.key, TaskInstanceState.FAILED)
        assert worker_with_job.free_concurrency == worker_with_job.concurrency

    @time_machine.travel(datetime.now(), tick=False)
    @patch("airflow.providers.edge3.cli.worker.logs_push")
    def test_check_running_jobs_log_push(self, mock_logs_push, worker_with_job: EdgeWorker):
        job = EdgeWorker.jobs[0]
        job.logfile.write_text("some log content")
        with conf_vars(
            {
                ("edge", "api_url"): "https://invalid-api-test-endpoint",
                ("edge", "push_log_chunk_size"): "524288",
            }
        ):
            worker_with_job.check_running_jobs()
        assert len(EdgeWorker.jobs) == 1
        mock_logs_push.assert_called_once_with(
            task=job.edge_job.key, log_chunk_time=timezone.utcnow(), log_chunk_data="some log content"
        )

    @time_machine.travel(datetime.now(), tick=False)
    @patch("airflow.providers.edge3.cli.worker.logs_push")
    def test_check_running_jobs_log_push_increment(self, mock_logs_push, worker_with_job: EdgeWorker):
        job = EdgeWorker.jobs[0]
        job.logfile.write_text("hello ")
        job.logsize = job.logfile.stat().st_size
        job.logfile.write_text("hello world")
        with conf_vars(
            {
                ("edge", "api_url"): "https://invalid-api-test-endpoint",
                ("edge", "push_log_chunk_size"): "524288",
            }
        ):
            worker_with_job.check_running_jobs()
        assert len(EdgeWorker.jobs) == 1
        mock_logs_push.assert_called_once_with(
            task=job.edge_job.key, log_chunk_time=timezone.utcnow(), log_chunk_data="world"
        )

    @time_machine.travel(datetime.now(), tick=False)
    @patch("airflow.providers.edge3.cli.worker.logs_push")
    def test_check_running_jobs_log_push_chunks(self, mock_logs_push, worker_with_job: EdgeWorker):
        job = EdgeWorker.jobs[0]
        job.logfile.write_bytes("log1log2ülog3".encode("latin-1"))
        with conf_vars(
            {("edge", "api_url"): "https://invalid-api-test-endpoint", ("edge", "push_log_chunk_size"): "4"}
        ):
            worker_with_job.check_running_jobs()
        assert len(EdgeWorker.jobs) == 1
        calls = mock_logs_push.call_args_list
        assert len(calls) == 4
        assert calls[0] == call(
            task=job.edge_job.key, log_chunk_time=timezone.utcnow(), log_chunk_data="log1"
        )
        assert calls[1] == call(
            task=job.edge_job.key, log_chunk_time=timezone.utcnow(), log_chunk_data="log2"
        )
        assert calls[2] == call(
            task=job.edge_job.key, log_chunk_time=timezone.utcnow(), log_chunk_data="\\xfc"
        )
        assert calls[3] == call(
            task=job.edge_job.key, log_chunk_time=timezone.utcnow(), log_chunk_data="log3"
        )

    @pytest.mark.parametrize(
        "drain, maintenance_mode, jobs, expected_state",
        [
            pytest.param(False, False, False, EdgeWorkerState.IDLE, id="idle"),
            pytest.param(False, False, True, EdgeWorkerState.RUNNING, id="running_jobs"),
            pytest.param(False, True, False, EdgeWorkerState.MAINTENANCE_MODE, id="maintenance_no_job"),
            pytest.param(
                False, True, True, EdgeWorkerState.MAINTENANCE_PENDING, id="maintenance_running_jobs"
            ),
            pytest.param(True, False, False, EdgeWorkerState.OFFLINE, id="shut_down"),
            pytest.param(True, False, True, EdgeWorkerState.TERMINATING, id="terminating"),
            pytest.param(True, True, False, EdgeWorkerState.OFFLINE_MAINTENANCE, id="offline_maintenance"),
            pytest.param(True, True, True, EdgeWorkerState.TERMINATING, id="maintenance_shut_down"),
        ],
    )
    @patch("airflow.providers.edge3.cli.worker.worker_set_state")
    def test_heartbeat(
        self, mock_set_state, drain, maintenance_mode, jobs, expected_state, worker_with_job: EdgeWorker
    ):
        if not jobs:
            EdgeWorker.jobs = []
        EdgeWorker.drain = drain
        EdgeWorker.maintenance_mode = maintenance_mode
        mock_set_state.return_value = WorkerSetStateReturn(
            state=EdgeWorkerState.RUNNING, queues=["queue1", "queue2"]
        )
        with conf_vars({("edge", "api_url"): "https://invalid-api-test-endpoint"}):
            worker_with_job.heartbeat()
        assert mock_set_state.call_args.args[1] == expected_state
        queue_list = worker_with_job.queues or []
        assert len(queue_list) == 2
        assert "queue1" in (queue_list)
        assert "queue2" in (queue_list)

    @patch("airflow.providers.edge3.cli.worker.worker_set_state")
    def test_version_mismatch(self, mock_set_state, worker_with_job):
        mock_set_state.side_effect = EdgeWorkerVersionException("")
        worker_with_job.heartbeat()
        assert worker_with_job.drain

    @patch("airflow.providers.edge3.cli.worker.worker_register")
    def test_start_missing_apiserver(self, mock_register_worker, worker_with_job: EdgeWorker):
        mock_response = Response()
        mock_response.status_code = 404
        mock_register_worker.side_effect = HTTPError(
            "Something with 404:NOT FOUND means API is not active", response=mock_response
        )
        with pytest.raises(SystemExit, match=r"API endpoint is not ready"):
            worker_with_job.start()

    @patch("airflow.providers.edge3.cli.worker.worker_register")
    def test_start_server_error(self, mock_register_worker, worker_with_job: EdgeWorker):
        mock_response = Response()
        mock_response.status_code = 500
        mock_register_worker.side_effect = HTTPError(
            "Something other error not FourhundretFour", response=mock_response
        )
        with pytest.raises(SystemExit, match=r"Something other"):
            worker_with_job.start()

    @patch("airflow.providers.edge3.cli.worker.worker_register")
    @patch("airflow.providers.edge3.cli.worker.EdgeWorker.loop")
    @patch("airflow.providers.edge3.cli.worker.worker_set_state")
    def test_start_and_run_one(self, mock_set_state, mock_loop, mock_register, worker_with_job: EdgeWorker):
        def stop_running():
            EdgeWorker.drain = True
            EdgeWorker.jobs = []

        mock_loop.side_effect = stop_running
        mock_register.side_effect = [WorkerRegistrationReturn(last_update=datetime.now())]

        worker_with_job.start()

        mock_register.assert_called_once()
        mock_loop.assert_called_once()
        assert mock_set_state.call_count == 2

    def test_get_sysinfo(self, worker_with_job: EdgeWorker):
        concurrency = 8
        worker_with_job.concurrency = concurrency
        sysinfo = worker_with_job._get_sysinfo()
        assert "airflow_version" in sysinfo
        assert "edge_provider_version" in sysinfo
        assert "concurrency" in sysinfo
        assert sysinfo["concurrency"] == concurrency

    @pytest.mark.db_test
    def test_list_edge_workers(self, mock_edgeworker: EdgeWorkerModel):
        args = self.parser.parse_args(["edge", "list-workers", "--output", "json"])
        with contextlib.redirect_stdout(StringIO()) as temp_stdout:
            with patch(
                "airflow.providers.edge3.models.edge_worker.get_registered_edge_hosts",
                return_value=[mock_edgeworker],
            ):
                edge_command.list_edge_workers(args)
                out = temp_stdout.getvalue()
                edge_workers = json.loads(out)
        for key in [
            "worker_name",
            "state",
            "queues",
            "jobs_active",
            "concurrency",
            "free_concurrency",
            "maintenance_comment",
        ]:
            assert key in edge_workers[0]
        assert any("test_edge_worker" in h["worker_name"] for h in edge_workers)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest
from requests import HTTPError
from requests.exceptions import ConnectTimeout
from requests_mock import ANY

from airflow.providers.edge3.cli.api_client import _make_generic_request

from tests_common.test_utils.config import conf_vars

if TYPE_CHECKING:
    from requests_mock import Mocker as RequestsMocker

MOCK_ENDPOINT = "https://invalid-api-test-endpoint"


class TestApiClient:
    @conf_vars({("edge", "api_url"): MOCK_ENDPOINT})
    def test_make_generic_request_success(self, requests_mock: RequestsMocker):
        requests_mock.get(
            ANY,
            [
                {"json": {"test": "ok"}},
                {"status_code": HTTPStatus.NO_CONTENT},
            ],
        )

        result1 = _make_generic_request("GET", f"{MOCK_ENDPOINT}/dummy_service", "test")
        result2 = _make_generic_request("GET", f"{MOCK_ENDPOINT}/service_no_content", "test")

        assert result1 == {"test": "ok"}
        assert result2 is None
        assert requests_mock.call_count == 2

    @patch("time.sleep", return_value=None)
    @conf_vars({("edge", "api_url"): MOCK_ENDPOINT})
    def test_make_generic_request_retry(self, mock_sleep, requests_mock: RequestsMocker):
        requests_mock.get(
            ANY,
            [
                *[{"status_code": HTTPStatus.SERVICE_UNAVAILABLE}] * 3,
                {"exc": ConnectTimeout},
                {"json": {"test": 42}},
            ],
        )

        result = _make_generic_request("GET", f"{MOCK_ENDPOINT}/flaky_service", "test")

        assert result == {"test": 42}
        assert requests_mock.call_count == 5

    @patch("time.sleep", return_value=None)
    @conf_vars({("edge", "api_url"): MOCK_ENDPOINT})
    def test_make_generic_request_unrecoverable_error(self, mock_sleep, requests_mock: RequestsMocker):
        requests_mock.get(
            ANY,
            [
                *[{"status_code": HTTPStatus.INTERNAL_SERVER_ERROR}] * 11,
                {"json": {"test": "uups"}},
            ],
        )

        with pytest.raises(HTTPError) as err:
            _make_generic_request("GET", f"{MOCK_ENDPOINT}/broken_service", "test")

        assert err.value.response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert requests_mock.call_count == 10
